/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.maven.plugins.buildconfigurator.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.el.ELContext;

import org.apache.commons.configuration.CompositeConfiguration;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.maven.execution.ExecutionEvent;
import org.apache.maven.execution.ExecutionListener;
import org.apache.maven.execution.MavenExecutionRequest;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.LegacySupport;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.component.annotations.Component;
import org.codehaus.plexus.component.annotations.Requirement;

import com.github.htfv.maven.plugins.buildconfigurator.el.BuildConfiguratorELContext;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCBuild;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCDependency;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCExclusion;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCExecution;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCExtensionModel;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCPlugin;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCPluginModel;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCProperties;
import com.github.htfv.maven.plugins.buildconfigurator.model.BCPropertyFile;
import com.github.htfv.maven.plugins.buildconfigurator.utils.ELUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils;

/**
 * The core logic of the Build Configurator.
 *
 * @author htfv (Aliaksei Lahachou)
 */
@Component(role = BuildConfigurator.class)
public class BuildConfigurator
{
    private static class ExecutionContext
    {
        public final boolean connectorRequest;
        public final ELContext elContext;
        public final BCExtensionModel extensionModel;
        public final CompositeConfiguration loadedProperties;
        public final Map<String, String> newProperties;
        public final BCPluginModel pluginModel;
        public final MavenProject project;

        public ExecutionContext(final ConfigureRequest request)
        {
            this.connectorRequest = request.isConnectorRequest();
            this.extensionModel   = request.getExtensionModel();
            this.loadedProperties = new CompositeConfiguration();
            this.newProperties    = new HashMap<String, String>();
            this.pluginModel      = request.getPluginModel();
            this.project          = request.getProject();

            this.elContext = BuildConfiguratorELContext
                    .newInstance(project, loadedProperties);
        }
    }

    /**
     * Default value for {@link BCDependency#getAction}.
     */
    private static final String DEFAULT_DEPENDENCY_ACTION = "add-update";

    /**
     * Default value for {@link BCDependency#getType}.
     */
    private static final String DEFAULT_DEPENDENCY_TYPE = "jar";

    /**
     * Default value for {@link BCExclusion#getAction}.
     */
    private static final String DEFAULT_EXCLUSION_ACTION = "add";

    /**
     * Default value for {@link BCExecution#getAction}.
     */
    private static final String DEFAULT_EXECUTION_ACTION = "add-update";

    /**
     * Default value for {@link BCExecution#getId}.
     */
    private static final String DEFAULT_EXECUTION_ID = "default";

    /**
     * Default value for {@link BCPlugin#getAction}.
     */
    private static final String DEFAULT_PLUGIN_ACTION = "add-update";

    /**
     * Default value for {@link BCProperties#getDefaultPropertyFilesEncoding}.
     */
    private static final String DEFAULT_PROPERTY_FILES_ENCODING = "ISO-8859-1";

    /**
     * Default value for {@link BCProperties#getIgnoreMissingPropertyFiles}.
     */
    private static final String DEFAULT_IGNORE_MISSING_PROPERTY_FILES = "false";

    /**
     * Valid values for {@link BCDependency#getAction}.
     */
    private static final String[] VALID_DEPENDENCY_ACTIONS =
            { "add", "add-update", "skip", "update" };

    /**
     * Valid values for {@link BCExclusion#getAction}.
     */
    private static final String[] VALID_EXCLUSION_ACTIONS =
            { "add", "skip" };

    /**
     * Valid values for {@link BCExecution#getAction}.
     */
    private static final String[] VALID_EXECUTION_ACTIONS =
            { "add", "add-update", "skip", "update" };

    /**
     * Valid values for {@link BCPlugin#getAction}.
     */
    private static final String[] VALID_PLUGIN_ACTIONS =
            { "add", "add-update", "skip", "update" };

    @Requirement
    private LegacySupport legacySupport;

    public ConfigureResult configure(final ConfigureRequest request)
            throws ConfigurationException, IOException
    {
        ExecutionContext ctx = new ExecutionContext(request);

        if (ctx.pluginModel != null)
        {
            normalizeMojoModel(ctx.pluginModel);
            configureMojoProject(ctx, ctx.pluginModel);
        }

        if (ctx.extensionModel != null)
        {
            normalizeExtensionModel(ctx.extensionModel);
            configureExtensionProject(ctx, ctx.extensionModel);
        }

        reinterpolateProjectModel(ctx);

        ConfigureResult result = new ConfigureResult.Builder()
                .newProperties(ctx.newProperties)
                .build();

        return result;
    }

    private void configureExtensionProject(final ExecutionContext ctx,
            final BCExtensionModel model) throws ConfigurationException, IOException
    {
        //
        // Update project plugins.
        //

        updatePlugins(ctx, model.getBuild().getPlugins());
    }

    private void configureMojoProject(final ExecutionContext ctx, final BCPluginModel model)
            throws ConfigurationException, IOException
    {
        //
        // Load property files.
        //

        loadPropertyFiles(ctx, model.getPropertyFiles());

        //
        // Update project properties with loaded values.
        //

        updateProperties(ctx);

        //
        // Update project dependencies.
        //

        if (updateDependencies(ctx, model.getDependencies()))
        {
            //
            // Reset project artifacts to make Maven resolve them again.
            //

            ctx.project.setDependencyArtifacts(null);
        }
    }

    /**
     * Loads single property file using parameters specified in the
     * {@code propertyFiles} element.
     *
     * @param project
     *            project for which property file shall be loaded.
     * @param propertyFile
     *            parameters for loading the property file.
     * @param elContext
     *            {@link ELContext} used to resolve EL expressions.
     *
     * @return A set of loaded properties, or {@code null} if file does not
     *         exist and {@code propertyFile.ignoreMissingFile} is {@code true}.
     *
     * @throws ConfigurationException
     *             if an property file could not be loaded.
     * @throws IOException
     *             if property file could not be read.
     * @throws PropertyFileMissingException
     *             if file does not exist and
     *             {@code propertyFile.ignoreMissingFile} is not {@code true}.
     */
    private Configuration loadPropertyFile(final MavenProject project,
            final BCPropertyFile propertyFile, final ELContext elContext)
            throws ConfigurationException, IOException
    {
        BufferedReader reader = null;

        try
        {
            //
            // Build absolute path name if relative file name was specified.
            //

            File file = new File(ELUtils.getStringValue(elContext, propertyFile.getFile()));

            if (!file.isAbsolute())
            {
                file = new File(project.getBasedir(), file.getPath());
            }

            //
            // Check to see whether the file exists.
            //

            if (!file.exists())
            {
                //
                // Throw an error if missing file is not allowed.
                //

                Boolean ignoreMissing = ELUtils.getBooleanValue(
                        elContext, propertyFile.getIgnoreMissing());

                if (ignoreMissing)
                {
                    return null;
                }
                else
                {
                    throw new PropertyFileMissingException(file);
                }
            }

            //
            // Open file reader with appropriate encoding.
            //

            String encoding = ELUtils.getStringValue(elContext, propertyFile.getEncoding());

            reader = new BufferedReader(new InputStreamReader(
                    new FileInputStream(file), encoding));

            //
            // And finally load the properties.
            //

            PropertiesConfiguration loadedPropertyFile = new PropertiesConfiguration();

            loadedPropertyFile.setDelimiterParsingDisabled(true);
            loadedPropertyFile.load(reader);

            return loadedPropertyFile;
        }
        finally
        {
            if (reader != null)
            {
                reader.close();
            }
        }
    }

    /**
     * Loads property files using parameters specified in the
     * {@code propertyFiles} element.
     *
     * @param project
     *            project for which property files shall be loaded.
     * @param propertyFiles
     *            parameters for loading property files.
     * @param elContext
     *            {@link ELContext} used to resolve EL expressions.
     * @param loadedPropertyFiles
     *            composite configuration to which the loaded property files
     *            shall be added.
     *
     * @throws ConfigurationException
     *             if an property file could not be loaded.
     * @throws IOException
     *             if property file could not be read.
     * @throws PropertyFileMissingException
     *             if file does not exist and
     *             {@code propertyFile.ignoreMissingFile} is not {@code true}.
     */
    private void loadPropertyFiles(final ExecutionContext ctx, final BCPropertyFile[] propertyFiles)
            throws ConfigurationException, IOException
    {
        for (int i = 0; i < propertyFiles.length; ++i)
        {
            Configuration loadedProperties = loadPropertyFile(
                    ctx.project, propertyFiles[i], ctx.elContext);

            if (loadedProperties != null)
            {
                ctx.loadedProperties.addConfiguration(loadedProperties);
            }
        }
    }

    /**
     * Initializes all uninitialized configuration parameters with default
     * values. This allows to avoid null-checks and simplifies the rest of the
     * code.
     *
     * @param model
     *            model to normalize.
     *
     * @return the model.
     *
     * @exception RequiredParameterNotSpecifiedException
     *                thrown if a required parameter was not specified.
     */
    private BCExtensionModel normalizeExtensionModel(final BCExtensionModel model)
    {
        normalizeMojoModel(model);

        //
        // build
        //

        BCBuild build = model.getBuild();

        if (build == null)
        {
            build = new BCBuild();
            model.setBuild(build);
        }

        //
        // build/plugins
        //

        BCPlugin[] plugins = build.getPlugins();

        if (plugins == null)
        {
            plugins = new BCPlugin[0];
            build.setPlugins(plugins);
        }

        //
        // build/plugins/plugin
        //

        normalizeModelPlugins(plugins);

        //
        // Return the normalized model.
        //

        return model;
    }

    /**
     * Initializes all uninitialized {@link BCDependency} properties with default values.
     *
     * @param dependencies
     *            {@link BCDependency} objects to normalize.
     * @param root
     *            path to the dependencies.
     *
     * @exception ParameterValueInvalidException
     *                throws if a parameter has an invalid value.
     * @exception RequiredParameterNotSpecifiedException
     *                thrown if a required parameter was not specified.
     */
    private void normalizeModelDependencies(final BCDependency[] dependencies, final String root)
    {
        for (BCDependency dependency : dependencies)
        {
            //
            // dependencies/dependency/action
            //

            if (StringUtils.isEmpty(dependency.getAction()))
            {
                dependency.setAction(DEFAULT_DEPENDENCY_ACTION);
            }
            else if (!ArrayUtils.contains(VALID_DEPENDENCY_ACTIONS, dependency.getAction()))
            {
                throw new ParameterValueInvalidException(
                        root + "dependencies/dependency/action");
            }

            //
            // dependencies/dependency/groupId
            //

            if (StringUtils.isEmpty(dependency.getGroupId()))
            {
                throw new RequiredParameterNotSpecifiedException(
                        root + "dependencies/dependency/groupId");
            }

            //
            // dependencies/dependency/artifactId
            //

            if (StringUtils.isEmpty(dependency.getArtifactId()))
            {
                throw new RequiredParameterNotSpecifiedException(
                        root + "dependencies/dependency/artifactId");
            }

            //
            // dependencies/dependency/type
            //

            if (dependency.getType() == null)
            {
                dependency.setType(DEFAULT_DEPENDENCY_TYPE);
            }

            //
            // dependencies/dependency/exclusions
            //

            if (dependency.getExclusions() == null)
            {
                dependency.setExclusions(new BCExclusion[0]);
            }

            //
            // dependencies/dependency/exclusions/exclusion
            //

            for (BCExclusion exclusion : dependency.getExclusions())
            {
                //
                // dependencies/dependency/exclusions/exclusion/action
                //

                if (StringUtils.isEmpty(exclusion.getAction()))
                {
                    exclusion.setAction(DEFAULT_EXCLUSION_ACTION);
                }
                else if (!ArrayUtils.contains(VALID_EXCLUSION_ACTIONS, exclusion.getAction()))
                {
                    throw new ParameterValueInvalidException(
                            root + "dependencies/dependency/exclusions/exclusion/action");
                }

                //
                // dependencies/dependency/exclusions/exclusion/groupId
                //

                if (StringUtils.isEmpty(exclusion.getGroupId()))
                {
                    throw new RequiredParameterNotSpecifiedException(
                            root + "dependencies/dependency/exclusions/exclusion/groupId");
                }

                //
                // dependencies/dependency/exclusions/exclusion/artifactId
                //

                if (StringUtils.isEmpty(exclusion.getArtifactId()))
                {
                    throw new RequiredParameterNotSpecifiedException(
                            root + "dependencies/dependency/exclusions/exclusion/artifactId");
                }
            }
        }
    }

    /**
     * Initializes all uninitialized {@link BCPlugin} properties with default values.
     *
     * @param plugins
     *            {@link BCPlugin} objects to normalize.
     *
     * @exception ParameterValueInvalidException
     *                throws if a parameter has an invalid value.
     * @exception RequiredParameterNotSpecifiedException
     *                thrown if a required parameter was not specified.
     */
    private void normalizeModelPlugins(final BCPlugin[] plugins)
    {
        for (BCPlugin plugin : plugins)
        {
            //
            // build/plugins/plugin/action
            //

            if (StringUtils.isEmpty(plugin.getAction()))
            {
                plugin.setAction(DEFAULT_PLUGIN_ACTION);
            }
            else if (!ArrayUtils.contains(VALID_PLUGIN_ACTIONS, plugin.getAction()))
            {
                throw new ParameterValueInvalidException("build/plugins/plugin/action");
            }

            //
            // build/plugins/plugin/groupId
            //
            // TODO: groupId is not mandatory in Maven, research how to deal
            //       with it.
            //

            if (StringUtils.isEmpty(plugin.getGroupId()))
            {
                throw new RequiredParameterNotSpecifiedException("build/plugins/plugin/groupId");
            }

            //
            // build/plugins/plugin/artifactId
            //

            if (StringUtils.isEmpty(plugin.getArtifactId()))
            {
                throw new RequiredParameterNotSpecifiedException("build/plugins/plugin/artifactId");
            }

            //
            // build/plugins/plugin/dependencies
            //

            BCDependency[] pluginDependencies = plugin.getDependencies();

            if (pluginDependencies == null)
            {
                pluginDependencies = new BCDependency[0];
                plugin.setDependencies(pluginDependencies);
            }

            //
            // build/plugins/plugin/dependencies/dependency
            //

            normalizeModelDependencies(pluginDependencies, "build/plugins/plugin/");

            //
            // build/plugins/plugin/executions
            //

            BCExecution[] executions = plugin.getExecutions();

            if (executions == null)
            {
                executions = new BCExecution[0];
                plugin.setExecutions(executions);
            }

            //
            // build/plugins/plugin/executions/execution
            //

            for (BCExecution execution : executions)
            {
                //
                // build/plugins/plugin/executions/execution/action
                //

                if (StringUtils.isEmpty(execution.getAction()))
                {
                    execution.setAction(DEFAULT_EXECUTION_ACTION);
                }
                else if (!ArrayUtils.contains(VALID_EXECUTION_ACTIONS, execution.getAction()))
                {
                    throw new ParameterValueInvalidException(
                            "build/plugins/plugin/executions/execution/action");
                }

                //
                // build/plugins/plugin/executions/execution/id
                //

                if (StringUtils.isEmpty(execution.getId()))
                {
                    execution.setId(DEFAULT_EXECUTION_ID);
                }

                //
                // build/plugins/plugin/executions/execution/goals
                //

                String[] goals = execution.getGoals();

                if (goals == null)
                {
                    goals = new String[0];
                    execution.setGoals(goals);
                }
            }
        }
    }

    private BCPluginModel normalizeMojoModel(final BCPluginModel model)
    {
        //
        // properties
        //

        BCProperties properties = model.getProperties();

        if (properties == null)
        {
            properties = new BCProperties();
            model.setProperties(properties);
        }

        //
        // properties/defaultPropertyFilesEncoding
        //

        if (properties.getDefaultPropertyFilesEncoding() == null)
        {
            properties.setDefaultPropertyFilesEncoding(DEFAULT_PROPERTY_FILES_ENCODING);
        }

        //
        // properties/ignoreMissingPropertyFiles
        //

        if (properties.getIgnoreMissingPropertyFiles() == null)
        {
            properties.setIgnoreMissingPropertyFiles(DEFAULT_IGNORE_MISSING_PROPERTY_FILES);
        }

        //
        // propertyFiles
        //

        BCPropertyFile[] propertyFiles = model.getPropertyFiles();

        if (propertyFiles == null)
        {
            propertyFiles = new BCPropertyFile[0];
            model.setPropertyFiles(propertyFiles);
        }

        //
        // propertyFiles/propertyFile
        //

        for (BCPropertyFile propertyFile : propertyFiles)
        {
            //
            // propertyFiles/propertyFile/encoding
            //

            if (propertyFile.getEncoding() == null)
            {
                propertyFile.setEncoding(properties.getDefaultPropertyFilesEncoding());
            }

            //
            // propertyFiles/propertyFile/file
            //

            if (propertyFile.getFile() == null)
            {
                throw new RequiredParameterNotSpecifiedException("propertyFiles/propertyFile/file");
            }

            //
            // propertyFiles/propertyFile/ignoreMissing
            //

            if (propertyFile.getIgnoreMissing() == null)
            {
                propertyFile.setIgnoreMissing(properties.getIgnoreMissingPropertyFiles());
            }
        }

        //
        // dependencies
        //

        BCDependency[] dependencies = model.getDependencies();

        if (dependencies == null)
        {
            dependencies = new BCDependency[0];
            model.setDependencies(dependencies);
        }

        //
        // dependencies/dependency
        //

        normalizeModelDependencies(dependencies, "");

        return model;
    }

    private void reinterpolateProjectModel(final ExecutionContext ctx)
    {
        //
        // Maven interpolates expressions when it loads the model. Since the
        // properties are loaded after the model, we have to interpolate
        // expressions again.
        //
        // It is actually not required for mojo parameters, because parameters
        // are interpolated before mojo is executed. However, it can be a
        // problem if mojo accesses model, like maven-resources-plugin accessing
        // project.build.resources.
        //
        // We take a conservative approach and only replace properties which
        // were loaded by the build configurator.
        //

        new ModelInterpolator(ctx.newProperties).interpolateModel(ctx.project.getModel());

        //
        // Maven interpolates mojo parameters before execution, but it skips
        // parameters of the PlexusConfiguration type. Updating the model now
        // has no effect, because execution plan is already calculated and mojo
        // configuration is copied to mojo execution. We cannot get access to
        // mojo execution, so we register a listener which will be call before
        // mojo is executed.
        //

        if (!ctx.connectorRequest)
        {
            final ModelInterpolator interpolator = new ModelInterpolator(ctx.newProperties);

            MavenSession session = legacySupport.getSession();
            MavenExecutionRequest executionRequest = session.getRequest();
            ExecutionListener executionListener = executionRequest.getExecutionListener();

            executionRequest.setExecutionListener(
                    new DelegatingExecutionListener(executionListener)
                    {
                        @Override
                        public void mojoStarted(final ExecutionEvent event)
                        {
                            interpolator.interpolateDOM(
                                    event.getMojoExecution().getConfiguration());

                            super.mojoStarted(event);
                        }

                        @Override
                        public void projectSucceeded(final ExecutionEvent event)
                        {
                            event.getSession().getRequest().setExecutionListener(getDelegate());
                            super.projectSucceeded(event);
                        }
                    });
        }
    }

    /**
     * Loads dependencies using parameters specified in the {@code dependencies}
     * element and updates project dependencies. The same algorithm for updating
     * dependencies is used as if they were in profile and profile were
     * activated.
     *
     * @param projectDependencies
     *            collection of dependencies which has to be updated.
     * @param dependencies
     *            parameters for updating dependencies.
     * @param elContext
     *            {@link ELContext} used to resolve EL expressions.
     *
     * @return {@code true} if the collection of dependencies was updated.
     */
    private boolean updateDependencies(final ExecutionContext ctx,
            final BCDependency[] dependencies)
    {
        return MergeUtils.mergeCollections(ctx.project.getDependencies(), dependencies,
                new DependencyMergeHelper(ctx.elContext));
    }

    /**
     * Uses parameters specified in the {@code build/plugins} element to update
     * project plugins. The algorithm used for updated is similar to the
     * algorithm used by Maven to merge plugin information from a profile when
     * it is activated.
     *
     * @param projectPlugins
     *            collection of plugins which shall be updated.
     * @param plugins
     *            parameters for updating the plugins.
     * @param elContext
     *            {@link ELContext} used to resolve EL expressions.
     */
    private void updatePlugins(final ExecutionContext ctx, final BCPlugin[] plugins)
    {
        MergeUtils.mergeCollections(ctx.project.getModel().getBuild().getPlugins(), plugins,
                new PluginMergeHelper(ctx.elContext));
    }

    /**
     * Updates project properties with values loaded from property files.
     *
     * @param projectProperties
     *            properties which shall be updated.
     * @param loadedPropertyFiles
     *            loaded property files.
     * @param elContext
     *            {@link ELContext} used to resolve EL expressions.
     */
    @SuppressWarnings("unchecked")
    private void updateProperties(final ExecutionContext ctx)
    {
        Iterator<String> keys = ctx.loadedProperties.getKeys();
        Properties projectProperties = ctx.project.getProperties();

        while (keys.hasNext())
        {
            String key = keys.next();

            if (projectProperties.containsKey(key))
            {
                continue;
            }

            String value = ELUtils.getStringValue(
                    ctx.elContext, ctx.loadedProperties.getString(key));

            projectProperties.setProperty(key, value);
            ctx.newProperties.put(key, value);
        }
    }
}
