/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.maven.plugins.buildconfigurator.core;

import javax.el.ELContext;

import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.maven.model.Dependency;

import com.github.htfv.maven.plugins.buildconfigurator.model.BCDependency;
import com.github.htfv.maven.plugins.buildconfigurator.utils.ELUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils.MergeHelper;

/**
 * {@link MergeHelper} which assists in dependency merging.
 *
 * @author htfv (Aliaksei Lahachou)
 */
class DependencyMergeHelper implements MergeHelper<Dependency, BCDependency>
{
    /**
     * {@link ELContext} used to evaluate EL expressions.
     */
    private final ELContext elContext;

    /**
     * Creates a new instance of {@link DependencyMergeHelper}.
     *
     * @param elContext
     *            {@link ELContext} used to evaluate EL expressions.
     */
    public DependencyMergeHelper(final ELContext elContext)
    {
        this.elContext = elContext;
    }

    /**
     * Converts the given source {@link BCDependency} object into a new target
     * {@link Dependency} object.
     *
     * @param dependency
     *            source {@link BCDependency} object.
     *
     * @return a new target {@link Dependency} object or {@code null} if
     *         {@link BCDependency#getAction} is {@code skip} or not {@code add}
     *         or {@code add-update}.
     */
    @Override
    public Dependency add(final BCDependency dependency)
    {
        String action = ELUtils.getStringValue(elContext, dependency.getAction());

        if (BCDependency.ACTION_SKIP.equals(action)
                || !BCDependency.ACTION_ADD.equals(action)
                        && !BCDependency.ACTION_ADD_UPDATE.equals(action))
        {
            return null;
        }

        Dependency projectDependency = new Dependency();

        projectDependency.setGroupId(ELUtils.getStringValue(
                elContext, dependency.getGroupId()));
        projectDependency.setArtifactId(ELUtils.getStringValue(
                elContext, dependency.getArtifactId()));
        projectDependency.setType(ELUtils.getStringValue(
                elContext, dependency.getType()));
        projectDependency.setClassifier(ELUtils.getStringValue(
                elContext, dependency.getClassifier()));

        return updateInternal(projectDependency, dependency);
    }

    /**
     * Compares the given target {@link Dependency} object with the given source
     * {@link BCDependency} object using {@code groupId, artifactId, type,} and
     * {@code classifier}.
     *
     * @param projectDependency
     *            target {@link Dependency} object.
     * @param dependency
     *            source {@link BCDependency} object.
     *
     * @return {@code true} if the given target and source objects may be
     *         considered equal.
     */
    @Override
    public boolean equals(final Dependency projectDependency, final BCDependency dependency)
    {
        String groupId = ELUtils.getStringValue(elContext, dependency.getGroupId());

        if (ObjectUtils.notEqual(projectDependency.getGroupId(), groupId))
        {
            return false;
        }

        String artifactId = ELUtils.getStringValue(elContext, dependency.getArtifactId());

        if (ObjectUtils.notEqual(projectDependency.getArtifactId(), artifactId))
        {
            return false;
        }

        String type = ELUtils.getStringValue(elContext, dependency.getType());

        if (ObjectUtils.notEqual(projectDependency.getType(), type))
        {
            return false;
        }

        String classifier = ELUtils.getStringValue(elContext, dependency.getClassifier());

        if (ObjectUtils.notEqual(projectDependency.getClassifier(), classifier))
        {
            return false;
        }

        return true;
    }

    /**
     * Updates the given target {@link Dependency} object with the given source
     * {@link BCDependency} object.
     *
     * @param projectDependency
     *            target {@link Dependency} object.
     * @param dependency
     *            source {@link BCDependency} object.
     *
     * @return updated target {@link Dependency} object or {@code null} if
     *         {@link BCDependency#getAction} is {@code skip} or not
     *         {@code add-update} or {@code update}.
     */
    @Override
    public Dependency update(final Dependency projectDependency, final BCDependency dependency)
    {
        String action = ELUtils.getStringValue(elContext, dependency.getAction());

        if (BCDependency.ACTION_SKIP.equals(action)
                || !BCDependency.ACTION_ADD_UPDATE.equals(action)
                        && !BCDependency.ACTION_UPDATE.equals(action))
        {
            return projectDependency;
        }

        return updateInternal(projectDependency, dependency);
    }

    /**
     * Updates all non-key properties of the given {@link Dependency} object and
     * merges contained dependency exclusions.
     *
     * @param projectDependency
     *            target {@link Dependency} object.
     * @param dependency
     *            source {@link BCDependency} object.
     *
     * @return updated target {@link Dependency} object.
     */
    public Dependency updateInternal(final Dependency projectDependency,
            final BCDependency dependency)
    {
        if (StringUtils.isNotEmpty(dependency.getVersion()))
        {
            projectDependency.setVersion(ELUtils.getStringValue(
                    elContext, dependency.getVersion()));
        }

        if (StringUtils.isNotEmpty(dependency.getScope()))
        {
            projectDependency.setScope(ELUtils.getStringValue(
                    elContext, dependency.getScope()));
        }

        if (StringUtils.isNotEmpty(dependency.getSystemPath()))
        {
            projectDependency.setSystemPath(ELUtils.getStringValue(
                    elContext, dependency.getSystemPath()));
        }

        if (StringUtils.isNotEmpty(dependency.getOptional()))
        {
            projectDependency.setOptional(ELUtils.getBooleanValue(
                    elContext, dependency.getOptional()));
        }

        MergeUtils.mergeCollections(
                projectDependency.getExclusions(), dependency.getExclusions(),
                new ExclusionMergeHelper(elContext));

        return projectDependency;
    }
}
