/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.maven.plugins.buildconfigurator.core;

import javax.el.ELContext;

import org.apache.commons.lang3.ObjectUtils;
import org.apache.maven.model.Exclusion;

import com.github.htfv.maven.plugins.buildconfigurator.model.BCExclusion;
import com.github.htfv.maven.plugins.buildconfigurator.utils.ELUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils.MergeHelper;

/**
 * {@link MergeHelper} which assists in dependency exclusion merging.
 *
 * @author htfv (Aliaksei Lahachou)
 */
class ExclusionMergeHelper implements MergeHelper<Exclusion, BCExclusion>
{
    /**
     * {@link ELContext} used to evaluate EL expressions.
     */
    private final ELContext elContext;

    /**
     * Creates a new instance of {@link ExclusionMergeHelper}.
     *
     * @param elContext
     *            {@link ELContext} used to evaluate EL expressions.
     */
    public ExclusionMergeHelper(final ELContext elContext)
    {
        this.elContext = elContext;
    }

    /**
     * Converts the given source {@link BCExclusion} object into a new target
     * {@link Exclusion} object.
     *
     * @param exclusion
     *            source {@link BCExclusion} object.
     *
     * @return a new target {@link Exclusion} object or {@code null} if
     *         {@link BCExclusion#getAction} is {@code skip}.
     */
    @Override
    public Exclusion add(final BCExclusion exclusion)
    {
        String action = ELUtils.getStringValue(elContext, exclusion.getAction());

        if (BCExclusion.ACTION_SKIP.equals(action))
        {
            return null;
        }

        Exclusion projectExclusion = new Exclusion();

        projectExclusion.setGroupId(ELUtils.getStringValue(
                elContext, exclusion.getGroupId()));
        projectExclusion.setArtifactId(ELUtils.getStringValue(
                elContext, exclusion.getArtifactId()));

        return projectExclusion;
    }

    /**
     * Compares the given target {@link Exclusion} object with the given source
     * {@link BCExclusion} object using {@code groupId} and {@code artifactId}.
     *
     * @param projectExclusion
     *            target {@link Exclusion} object.
     * @param exclusion
     *            source {@link BCExclusion} object.
     *
     * @return {@code true} if the given target and source objects may be
     *         considered equal.
     */
    @Override
    public boolean equals(final Exclusion projectExclusion, final BCExclusion exclusion)
    {
        String groupId = ELUtils.getStringValue(elContext, exclusion.getGroupId());

        if (ObjectUtils.notEqual(projectExclusion.getGroupId(), groupId))
        {
            return false;
        }

        String artifactId = ELUtils.getStringValue(elContext, exclusion.getArtifactId());

        if (ObjectUtils.notEqual(projectExclusion.getArtifactId(), artifactId))
        {
            return false;
        }

        return true;
    }

    /**
     * Updates the given target {@link Exclusion} object with the given source
     * {@link BCExclusion} object.
     *
     * @param projectExclusion
     *            target {@link Exclusion} object.
     * @param exclusion
     *            source {@link BCExclusion} object.
     *
     * @return updated target {@link Exclusion} object.
     */
    @Override
    public Exclusion update(final Exclusion projectExclusion, final BCExclusion exclusion)
    {
        return projectExclusion;
    }
}
