/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.maven.plugins.buildconfigurator.core;

import javax.el.ELContext;

import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.maven.model.PluginExecution;

import com.github.htfv.maven.plugins.buildconfigurator.model.BCExecution;
import com.github.htfv.maven.plugins.buildconfigurator.utils.ELUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils.MergeHelper;

/**
 * {@link MergeHelper} which assists in plugin execution merging.
 *
 * @author htfv (Aliaksei Lahachou)
 */
class ExecutionMergeHelper implements MergeHelper<PluginExecution, BCExecution>
{
    /**
     * {@link ELContext} used to evaluate EL expressions.
     */
    private final ELContext elContext;

    /**
     * Creates a new instance of {@link ExecutionMergeHelper}.
     *
     * @param elContext
     *            {@link ELContext} used to evaluate EL expressions.
     */
    public ExecutionMergeHelper(final ELContext elContext)
    {
        this.elContext = elContext;
    }

    /**
     * Converts the given source {@link BCExecution} object into a new target
     * {@link PluginExecution} object.
     *
     * @param execution
     *            source {@link BCExecution} object.
     *
     * @return a new target {@link PluginExecution} object or {@code null} if
     *         {@link BCExecution#getAction} is {@code skip} or not {@code add}
     *         or {@code add-update}.
     */
    @Override
    public PluginExecution add(final BCExecution execution)
    {
        String action = ELUtils.getStringValue(elContext, execution.getAction());

        if (BCExecution.ACTION_SKIP.equals(action)
                || !BCExecution.ACTION_ADD.equals(action)
                        && !BCExecution.ACTION_ADD_UPDATE.equals(action))
        {
            return null;
        }

        PluginExecution projectExecution = new PluginExecution();

        projectExecution.setId(ELUtils.getStringValue(elContext, execution.getId()));

        return updateInternal(projectExecution, execution);
    }

    /**
     * Compares the given target {@link PluginExecution} object with the given
     * source {@link BCExecution} object using {@code id}.
     *
     * @param projectExecution
     *            target {@link PluginExecution} object.
     * @param execution
     *            source {@link BCExecution} object.
     *
     * @return {@code true} if the given target and source objects may be
     *         considered equal.
     */
    @Override
    public boolean equals(final PluginExecution projectExecution, final BCExecution execution)
    {
        String id = ELUtils.getStringValue(elContext, execution.getId());

        if (ObjectUtils.notEqual(projectExecution.getId(), id))
        {
            return false;
        }

        return true;
    }

    /**
     * Updates the given target {@link PluginExecution} object with the given
     * source {@link BCExecution} object.
     *
     * @param projectExecution
     *            target {@link PluginExecution} object.
     * @param execution
     *            source {@link BCExecution} object.
     *
     * @return updated target {@link PluginExecution} object or {@code null} if
     *         {@link BCExecution#getAction} is {@code skip} or not
     *         {@code add-update} or {@code update}.
     */
    @Override
    public PluginExecution update(final PluginExecution projectExecution,
            final BCExecution execution)
    {
        String action = ELUtils.getStringValue(elContext, execution.getAction());

        if (BCExecution.ACTION_SKIP.equals(action)
                || !BCExecution.ACTION_ADD_UPDATE.equals(action)
                        && !BCExecution.ACTION_UPDATE.equals(action))
        {
            return projectExecution;
        }

        return updateInternal(projectExecution, execution);
    }

    /**
     * Updates all non-key properties of the given {@link PluginExecution}
     * object and merges contained configuration and goals.
     *
     * @param projectExecution
     *            target {@link PluginExecution} object.
     * @param execution
     *            source {@link BCExecution} object.
     *
     * @return updated target {@link PluginExecution} object.
     */
    public PluginExecution updateInternal(final PluginExecution projectExecution,
            final BCExecution execution)
    {
        projectExecution.setConfiguration(
                MergeUtils.mergeConfiguration(
                        projectExecution.getConfiguration(), execution.getConfiguration(),
                        elContext));

        MergeUtils.mergeCollections(
                projectExecution.getGoals(), execution.getGoals(),
                new StringMergeHelper(elContext));

        if (StringUtils.isNotEmpty(execution.getPhase()))
        {
            projectExecution.setPhase(ELUtils.getStringValue(
                    elContext, execution.getPhase()));
        }

        return projectExecution;
    }
}
