package com.github.htfv.maven.plugins.buildconfigurator.core;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.apache.maven.model.Build;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.Exclusion;
import org.apache.maven.model.Extension;
import org.apache.maven.model.Model;
import org.apache.maven.model.Plugin;
import org.apache.maven.model.PluginExecution;
import org.apache.maven.model.PluginManagement;
import org.apache.maven.model.Resource;
import org.codehaus.plexus.interpolation.InterpolationException;
import org.codehaus.plexus.interpolation.Interpolator;
import org.codehaus.plexus.interpolation.StringSearchInterpolator;
import org.codehaus.plexus.util.xml.Xpp3Dom;

public class ModelInterpolator
{
    private final Interpolator interpolator;

    public ModelInterpolator(final Map<String, String> properties)
    {
        this.interpolator = new StringSearchInterpolator();

        interpolator.addValueSource(new PropertiesValueSource(properties));
        interpolator.setCacheAnswers(true);
    }

    private void interpolateBuild(final Build build)
    {
        build.setDefaultGoal(
                interpolateString(build.getDefaultGoal()));
        build.setDirectory(
                interpolateString(build.getDirectory()));
        build.setFinalName(
                interpolateString(build.getFinalName()));
        build.setOutputDirectory(
                interpolateString(build.getOutputDirectory()));
        build.setScriptSourceDirectory(
                interpolateString(build.getScriptSourceDirectory()));
        build.setSourceDirectory(
                interpolateString(build.getSourceDirectory()));
        build.setTestOutputDirectory(
                interpolateString(build.getTestOutputDirectory()));
        build.setTestSourceDirectory(
                interpolateString(build.getTestSourceDirectory()));

        interpolateExtensions(build.getExtensions());
        interpolateStrings(build.getFilters());
        interpolatePluginManagement(build.getPluginManagement());
        interpolatePlugins(build.getPlugins());
        interpolateResources(build.getResources());
        interpolateResources(build.getTestResources());
    }

    private void interpolateDependencies(final Collection<Dependency> dependencies)
    {
        for (Dependency dependency : dependencies)
        {
            interpolateDependency(dependency);
        }
    }

    private void interpolateDependency(final Dependency dependency)
    {
        dependency.setArtifactId(
                interpolateString(dependency.getArtifactId()));
        dependency.setClassifier(
                interpolateString(dependency.getClassifier()));
        dependency.setGroupId(
                interpolateString(dependency.getGroupId()));
        dependency.setOptional(
                interpolateString(dependency.getOptional()));
        dependency.setScope(
                interpolateString(dependency.getScope()));
        dependency.setSystemPath(
                interpolateString(dependency.getSystemPath()));
        dependency.setType(
                interpolateString(dependency.getType()));
        dependency.setVersion(
                interpolateString(dependency.getVersion()));

        interpolateExclusions(dependency.getExclusions());
    }

    public void interpolateDOM(final Xpp3Dom dom)
    {
        if (dom == null)
        {
            return;
        }

        dom.setValue(interpolateString(dom.getValue()));

        for (Xpp3Dom child : dom.getChildren())
        {
            interpolateDOM(child);
        }
    }

    private void interpolateExclusion(final Exclusion exclusion)
    {
        exclusion.setArtifactId(
                interpolateString(exclusion.getArtifactId()));
        exclusion.setGroupId(
                interpolateString(exclusion.getGroupId()));
    }

    private void interpolateExclusions(final Collection<Exclusion> exclusions)
    {
        for (Exclusion exclusion : exclusions)
        {
            interpolateExclusion(exclusion);
        }
    }

    private void interpolateExecution(final PluginExecution execution)
    {
        execution.setId(
                interpolateString(execution.getId()));
        execution.setInherited(
                interpolateString(execution.getInherited()));
        execution.setPhase(
                interpolateString(execution.getPhase()));

        interpolateDOM((Xpp3Dom) execution.getConfiguration());
        interpolateStrings(execution.getGoals());
    }

    private void interpolateExecutions(final Collection<PluginExecution> executions)
    {
        for (PluginExecution execution : executions)
        {
            interpolateExecution(execution);
        }
    }

    private void interpolateExtension(final Extension extension)
    {
        extension.setArtifactId(
                interpolateString(extension.getArtifactId()));
        extension.setGroupId(
                interpolateString(extension.getGroupId()));
        extension.setVersion(
                interpolateString(extension.getVersion()));
    }

    private void interpolateExtensions(final Collection<Extension> extensions)
    {
        for (Extension extension : extensions)
        {
            interpolateExtension(extension);
        }
    }

    public void interpolateModel(final Model model)
    {
        interpolateBuild(model.getBuild());
    }

    private void interpolatePlugin(final Plugin plugin)
    {
        plugin.setArtifactId(
                interpolateString(plugin.getArtifactId()));
        plugin.setExtensions(
                interpolateString(plugin.getExtensions()));
        plugin.setGroupId(
                interpolateString(plugin.getGroupId()));
        plugin.setInherited(
                interpolateString(plugin.getInherited()));
        plugin.setVersion(
                interpolateString(plugin.getVersion()));

        interpolateDOM((Xpp3Dom) plugin.getConfiguration());
        interpolateDependencies(plugin.getDependencies());
        interpolateExecutions(plugin.getExecutions());
    }

    private void interpolatePluginManagement(final PluginManagement pluginManagement)
    {
        interpolatePlugins(pluginManagement.getPlugins());
    }

    public void interpolatePlugins(final Collection<Plugin> plugins)
    {
        for (Plugin plugin : plugins)
        {
            interpolatePlugin(plugin);
        }
    }

    private void interpolateResource(final Resource resource)
    {
        resource.setDirectory(interpolateString(resource.getDirectory()));
    }

    private void interpolateResources(final Collection<Resource> resources)
    {
        for (Resource resource : resources)
        {
            interpolateResource(resource);
        }
    }

    private String interpolateString(final String string)
    {
        if (StringUtils.isEmpty(string))
        {
            return string;
        }

        try
        {
            return interpolator.interpolate(string);
        }
        catch (InterpolationException e)
        {
            return string;
        }
    }

    private void interpolateStrings(final List<String> filters)
    {
        for (int i = 0; i < filters.size(); i++)
        {
            filters.set(i, interpolateString(filters.get(i)));
        }
    }
}
