/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.maven.plugins.buildconfigurator.core;

import javax.el.ELContext;

import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.maven.model.Plugin;

import com.github.htfv.maven.plugins.buildconfigurator.model.BCPlugin;
import com.github.htfv.maven.plugins.buildconfigurator.utils.ELUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils;
import com.github.htfv.maven.plugins.buildconfigurator.utils.MergeUtils.MergeHelper;

/**
 * {@link MergeHelper} which assists in plugin merging.
 *
 * @author htfv (Aliaksei Lahachou)
 */
class PluginMergeHelper implements MergeHelper<Plugin, BCPlugin>
{
    /**
     * {@link ELContext} used to evaluate EL expressions.
     */
    private final ELContext elContext;

    /**
     * Creates a new instance of {@link PluginMergeHelper}.
     *
     * @param elContext
     *            {@link ELContext} used to evaluate EL expressions.
     */
    public PluginMergeHelper(final ELContext elContext)
    {
        this.elContext = elContext;
    }

    /**
     * Converts the given source {@link BCPlugin} object into a new target
     * {@link Plugin} object.
     *
     * @param plugin
     *            source {@link BCPlugin} object.
     *
     * @return a new target {@link Plugin} object or {@code null} if
     *         {@link BCPlugin#getAction} is {@code skip} or not {@code add} or
     *         {@code add-update}.
     */
    @Override
    public Plugin add(final BCPlugin plugin)
    {
        String action = ELUtils.getStringValue(elContext, plugin.getAction());

        if (BCPlugin.ACTION_SKIP.equals(action)
                || !BCPlugin.ACTION_ADD.equals(action)
                        && !BCPlugin.ACTION_ADD_UPDATE.equals(action))
        {
            return null;
        }

        Plugin projectPlugin = new Plugin();

        projectPlugin.setGroupId(ELUtils.getStringValue(
                elContext, plugin.getGroupId()));
        projectPlugin.setArtifactId(ELUtils.getStringValue(
                elContext, plugin.getArtifactId()));

        return updateInternal(projectPlugin, plugin);
    }

    /**
     * Compares the given target {@link Plugin} object with the given source
     * {@link BCPlugin} object using {@code groupId} and {@code artifactId}.
     *
     * @param projectPlugin
     *            target {@link Plugin} object.
     * @param plugin
     *            source {@link BCPlugin} object.
     *
     * @return {@code true} if the given target and source objects may be
     *         considered equal.
     */
    @Override
    public boolean equals(final Plugin projectPlugin, final BCPlugin plugin)
    {
        String groupId = ELUtils.getStringValue(elContext, plugin.getGroupId());

        if (ObjectUtils.notEqual(projectPlugin.getGroupId(), groupId))
        {
            return false;
        }

        String artifactId = ELUtils.getStringValue(elContext, plugin.getArtifactId());

        if (ObjectUtils.notEqual(projectPlugin.getArtifactId(), artifactId))
        {
            return false;
        }

        return true;
    }

    /**
     * Updates the given target {@link Plugin} object with the given source
     * {@link BCPlugin} object.
     *
     * @param projectPlugin
     *            target {@link Plugin} object.
     * @param plugin
     *            source {@link BCPlugin} object.
     *
     * @return updated target {@link Plugin} object or {@code null} if
     *         {@link BCPlugin#getAction} is {@code skip} or not
     *         {@code add-update} or {@code update}.
     */
    @Override
    public Plugin update(final Plugin projectPlugin, final BCPlugin plugin)
    {
        String action = ELUtils.getStringValue(elContext, plugin.getAction());

        if (BCPlugin.ACTION_SKIP.equals(action)
                || !BCPlugin.ACTION_ADD_UPDATE.equals(action)
                        && !BCPlugin.ACTION_UPDATE.equals(action))
        {
            return projectPlugin;
        }

        return updateInternal(projectPlugin, plugin);
    }

    /**
     * Updates all non-key properties of the given {@link Plugin} object and
     * merges contained configuration, dependencies, and executions.
     *
     * @param projectPlugin
     *            target {@link Plugin} object.
     * @param plugin
     *            source {@link BCPlugin} object.
     *
     * @return updated target {@link Plugin} object.
     */
    private Plugin updateInternal(final Plugin projectPlugin, final BCPlugin plugin)
    {
        if (StringUtils.isNotEmpty(plugin.getVersion()))
        {
            projectPlugin.setVersion(ELUtils.getStringValue(
                    elContext, plugin.getVersion()));
        }

        projectPlugin.setConfiguration(
                MergeUtils.mergeConfiguration(
                        projectPlugin.getConfiguration(), plugin.getConfiguration(),
                        elContext));

        MergeUtils.mergeCollections(
                projectPlugin.getDependencies(), plugin.getDependencies(),
                new DependencyMergeHelper(elContext));

        if (MergeUtils.mergeCollections(
                projectPlugin.getExecutions(), plugin.getExecutions(),
                new ExecutionMergeHelper(elContext)))
        {
            projectPlugin.flushExecutionMap();
        }

        return projectPlugin;
    }
}
