/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.maven.plugins.buildconfigurator.model;

import org.apache.commons.lang3.ArrayUtils;
import org.codehaus.plexus.configuration.PlexusConfiguration;

/**
 * Contains parameters for updating a plugin.
 *
 * @author htfv (Aliaksei Lahachou)
 */
public class BCPlugin
{
    /**
     * Plugin is added if it's not yet defined.
     */
    public static final String ACTION_ADD = "add";

    /**
     * Plugin is added if it's not yet defined, or updated if it's already
     * defined.
     */
    public static final String ACTION_ADD_UPDATE = "add-update";

    /**
     * Plugin is skipped.
     */
    public static final String ACTION_SKIP = "skip";

    /**
     * Plugin is updated if it's already defined.
     */
    public static final String ACTION_UPDATE = "update";

    /**
     * Action to perform with this plugin; may take one of the following values:
     * {@link BCPlugin#ACTION_ADD}, {@link BCPlugin#ACTION_ADD_UPDATE},
     * {@link BCPlugin#ACTION_SKIP}, {@link BCPlugin#ACTION_UPDATE}.
     */
    private String action;

    /**
     * {@code artifactId} of the plugin to update.
     */
    private String artifactId;

    /**
     * Plugin configuration.
     */
    private PlexusConfiguration configuration;

    /**
     * Plugin dependencies.
     */
    private BCDependency[] dependencies;

    /**
     * Plugin executions.
     */
    private BCExecution[] executions;

    /**
     * {@code groupId} of the plugin to update.
     */
    private String groupId;

    /**
     * {@code version} of the plugin to update.
     */
    private String version;

    /**
     * Gets the action to perform with this plugin; may take one of the
     * following values: {@link BCPlugin#ACTION_ADD},
     * {@link BCPlugin#ACTION_ADD_UPDATE}, {@link BCPlugin#ACTION_SKIP},
     * {@link BCPlugin#ACTION_UPDATE}.
     *
     * @return action to perform.
     */
    public String getAction()
    {
        return action;
    }

    /**
     * Gets {@code artifactId} of the plugin to update.
     *
     * @return {@code artifactId} of the plugin to update.
     */
    public String getArtifactId()
    {
        return artifactId;
    }

    /**
     * Gets plugin configuration.
     *
     * @return plugin configuration.
     */
    public PlexusConfiguration getConfiguration()
    {
        return configuration;
    }

    /**
     * Gets plugin dependencies.
     *
     * @return plugin dependencies.
     */
    public BCDependency[] getDependencies()
    {
        return ArrayUtils.clone(dependencies);
    }

    /**
     * Gets plugin executions.
     *
     * @return plugin executions.
     */
    public BCExecution[] getExecutions()
    {
        return ArrayUtils.clone(executions);
    }

    /**
     * Gets {@code groupId} of the plugin to update.
     *
     * @return {@code groupId} of the plugin to update.
     */
    public String getGroupId()
    {
        return groupId;
    }

    /**
     * Gets {@code version} of the plugin to update.
     *
     * @return {@code version} of the plugin to update.
     */
    public String getVersion()
    {
        return version;
    }

    /**
     * Sets the action to perform with this plugin; may take one of the
     * following values: {@link BCPlugin#ACTION_ADD},
     * {@link BCPlugin#ACTION_ADD_UPDATE}, {@link BCPlugin#ACTION_SKIP},
     * {@link BCPlugin#ACTION_UPDATE}.
     *
     * @param action
     *            action to perform.
     */
    public void setAction(final String action)
    {
        this.action = action;
    }

    /**
     * Sets {@code artifactId} of the plugin to update.
     *
     * @param artifactId
     *            {@code artifactId} of the plugin to update.
     */
    public void setArtifactId(final String artifactId)
    {
        this.artifactId = artifactId;
    }

    /**
     * Sets plugin configuration.
     *
     * @param configuration
     *            plugin configuration.
     */
    public void setConfiguration(final PlexusConfiguration configuration)
    {
        this.configuration = configuration;
    }

    /**
     * Sets plugin dependencies.
     *
     * @param dependencies
     *            plugin dependencies.
     */
    public void setDependencies(final BCDependency[] dependencies)
    {
        this.dependencies = ArrayUtils.clone(dependencies);
    }

    /**
     * Sets plugin executions.
     *
     * @param executions
     *            plugin executions.
     */
    public void setExecutions(final BCExecution[] executions)
    {
        this.executions = ArrayUtils.clone(executions);
    }

    /**
     * Sets {@code groupId} of the plugin to update.
     *
     * @param groupId
     *            {@code groupId} of the plugin to update.
     */
    public void setGroupId(final String groupId)
    {
        this.groupId = groupId;
    }

    /**
     * Sets {@code version} of the plugin to update.
     *
     * @param version
     *            {@code version} of the plugin to update.
     */
    public void setVersion(final String version)
    {
        this.version = version;
    }
}
